NAME		gnuplot
DESCRIPTION	'gnuplot plotting program'
EXETYPE		WINDOWS
SEGMENTS
	WINMAIN_TEXT   MOVEABLE NONDISCARDABLE PRELOAD
	COMMAND_TEXT   MOVEABLE NONDISCARDABLE PRELOAD
	PLOT_TEXT      MOVEABLE NONDISCARDABLE PRELOAD
	MISC_TEXT      MOVEABLE NONDISCARDABLE PRELOAD
	READLINE_TEXT  MOVEABLE NONDISCARDABLE PRELOAD
CODE		MOVEABLE DISCARDABLE PRELOAD
DATA		MULTIPLE MOVEABLE PRELOAD
HEAPSIZE	256
STACKSIZE	16000
